/* ReadTheDocs theme colors */
.wy-nav-top {
	background-color: #404040 
}
.wy-nav-content {
	max-width: 950px 
}
.wy-side-nav-search {
	background-color: transparent 
}
.wy-side-nav-search input[type="text"] {
	border-width: 0 
}
/* Custom classes */
.small {
	font-size:40% 
}
.smaller, .pr, .noteversion {
	font-size:70% 
}
.noteversion::after {
	content: "/" 
}
/* Code: literals and links */
.rst-content tt.literal,.rst-content code.literal {
	color: #404040;
}
/* slim font weight for non-link code */
.rst-content tt:not(.xref),.rst-content code:not(.xref),.rst-content *:not(a) > tt.xref,.rst-content *:not(a) > code.xref {
	font-weight: normal;
}
/* Just one box for annotation code for a less noisy look */
.rst-content .annotation {
	padding: 2px 5px;
	background-color: white;
	border: 1px solid #e1e4e5;
}
.rst-content .annotation tt,.rst-content .annotation code {
	padding: 0 0;
	background-color: transparent;
	border: 0 solid transparent;
}
/* Parameter lists */
/* Mimick rubric style used for other headings */
.rst-content dl:not(.docutils) dl > dt {
	font-weight: bold;
	background: none transparent;
	border-left: none;
	margin: 0 0 12px;
	padding: 3px 0 0;
	font-size: 105%;
}
html.writer-html5 .rst-content dl[class]:not(.option-list):not(.field-list):not(.footnote):not(.glossary):not(.simple) > dt {
	color: #404040;
	border-top: solid 3px #343131;
	background: #edf0f2;
}
/*class="longtable docutils align-default"*/
table.longtable td {
	white-space: normal!important;
}
/* Parameters contain <strong> parts and don’t need bold font */
.rst-content dl.field-list dl > dt {
	font-weight: unset!important;
}
div.highlight > pre {
    background: #F5F5F5;
}
/* Add colon between return tuple element name and type */
.rst-content dl:not(.docutils) dl > dt .classifier::before {
	content: ' : ' 
}
.rst-footer-buttons {
	margin: 24px 0px 0px 0px;
}
/* prolog bg */
.rst-content .note {
	background: #FFC76E;
}
.rst-content .output_area img {
	max-width: unset;
	width: 100% !important;
	height: auto !important;
}
.rst-content div[class^="highlight"] pre {
	white-space: pre;
	margin: 0;
	padding: 5px;
	display: block;
	overflow: auto;
}
/* sphinx-gallery */
div.sphx-glr-download a {
	background-color: #FFC76E !important;
	background-image: none !important;
	border-radius: 2px !important;
	border: 1px solid #f49600 !important;
	color: #404040 !important;
	font-weight bold !important;
	padding: 0.1cm !important;
	text-align: center !important;
}
div.sphx-glr-download a[href$=".py"] {
	display: none !important;
}
div.sphx-glr-example-title div[class="highlight"],
div.nbinput.container div.input_area > div[class^="highlight"] {
    background-color: #F5F5F5!important;
    border: none!important;
}
.sphx-glr-script-out .highlight pre {
	background: #FCFCFC !important;
}
p.sphx-glr-script-out {
	display: none !important;
}
div.sphx-glr-download p {
	margin: 0 !important;
	width: auto !important;
}
.sphx-glr-script-out {
	color: #404040 !important;
	margin: -24px 0px 0px 0px !important;
}
p.sphx-glr-signature {
	display: none !important;
}
div.sphx-glr-download-link-note {
	display: none !important;
}
/* this gets rid of uneven vertical padding */
div.sphx-glr-download code.download {
	display: block !important;
}
.sphx-glr-thumbcontainer {
	background: none !important;
	border: 1px solid #f49600 !important;
	text-align: center !important;
	min-height: 240px !important;
}
.sphx-glr-thumbcontainer .headerlink {
	display: none !important;
}
div.sphx-glr-thumbcontainer span {
	font-style: normal !important;
}
p.sphx-glr-timing {
	margin: 0 !important;
	padding-top: 24px;
	border-top: 1px solid #000;
}
.sphx-glr-thumbcontainer:hover {
	box-shadow: 0 0 15px RGBA(244, 150, 0, 0.5) !important;
}
/* sphinx-gallery inserts 2 <br> after_repr_html_, ignore the 1st one */
div[class="rendered_html"] + br {
	display: none!important;
}
span[id^="sphx-glr-auto-examples"] {
	min-height: 24px !important;
	display: block !important;
}
